// ------------------------------------------------------------------------------------
//
// Copyright  2006 Intel Corporation
// All Rights Reserved
// 
// The sample source code contained or described herein and all documents
// related to the sample source code ("Material") are owned by Intel Corporation
// or its suppliers or licensors. Title to the Material remains with Intel Corporation
// or its suppliers and licensors. The Material may contain trade secrets and
// proprietary and confidential information of Intel Corporation and its suppliers
// and licensors, and is protected by worldwide copyright and trade secret laws and
// treaty provisions. The sample source code is provided AS IS, with no warranties
// of any kind, express or implied. Except as expressly permitted by the Software
// license, neither Intel Corporation nor its suppliers assumes any responsibility
// or liability for any errors or inaccuracies that may appear herein.
//
// ------------------------------------------------------------------------------------
#include "stdafx.h"
#include "CriticalSection.h"

/** \brief constructor

This is the public constructor for CriticalSection.
Call isValid() to check if the object was created successfully.
\sa isValid()
*/
CriticalSection::CriticalSection(void)
{
    m_valid = false;
    #ifdef WIN32
        InitializeCriticalSection(&m_criticalSection);
        m_valid = true;
    #elif LINUX
        m_criticalSection = (pthread_mutex_t *)malloc(sizeof(pthread_mutex_t));
        if(m_criticalSection != NULL)
        { 
            if(pthread_mutex_init(m_criticalSection, NULL) == 0)
            {
                m_valid = true;
            }
        }
    #endif
}

/** \brief destructor

This is the public destructor for CriticalSection.
*/
CriticalSection::~CriticalSection(void)
{
    #ifdef WIN32
        DeleteCriticalSection(&m_criticalSection);
    #elif LINUX
        if(m_criticalSection != NULL)
        {
            pthread_mutex_destroy(m_criticalSection);
            free(m_criticalSection);
        }
    #endif
    m_valid = false;
}

/** \brief determine if the CriticalSection object is valid

Call this function to determine if the CriticalSection object was constructed successfully.
\return true if the object is valid, false otherwise
*/
BOOL CriticalSection::isValid(void)
{
    return m_valid;
}

/** \brief enters a critical section

Call this function to block the current thread until it has exclusive access to the code following the call.
Be sure to call leave() when exclusive access is no longer needed.
\sa leave()
*/
BOOL CriticalSection::enter(void)
{
    ASSERT(m_valid);
    
    BOOL output = false;
    
    #ifdef WIN32
        EnterCriticalSection(&m_criticalSection);
        output = true;
    #elif LINUX
        if(pthread_mutex_lock(m_criticalSection) == 0)
        {
            output = true;
        }
    #endif
    
    return output;
}

/** \brief leaves a critical section

Call this function to relinquish exclusive access to a region of code, which began with a call
to enter().
\sa enter()
*/
BOOL CriticalSection::leave(void)
{
    ASSERT(m_valid);
    
    BOOL output = false;

    #ifdef WIN32
        LeaveCriticalSection(&m_criticalSection);
        output = true;
    #elif LINUX
        if(pthread_mutex_unlock(m_criticalSection) == 0)
        {
            output = true;
        }
    #endif
    
    return output;
}
