// ------------------------------------------------------------------------------------
//
// Copyright  2006 Intel Corporation
// All Rights Reserved
// 
// The sample source code contained or described herein and all documents
// related to the sample source code ("Material") are owned by Intel Corporation
// or its suppliers or licensors. Title to the Material remains with Intel Corporation
// or its suppliers and licensors. The Material may contain trade secrets and
// proprietary and confidential information of Intel Corporation and its suppliers
// and licensors, and is protected by worldwide copyright and trade secret laws and
// treaty provisions. The sample source code is provided AS IS, with no warranties
// of any kind, express or implied. Except as expressly permitted by the Software
// license, neither Intel Corporation nor its suppliers assumes any responsibility
// or liability for any errors or inaccuracies that may appear herein.
//
// ------------------------------------------------------------------------------------
#include "stdafx.h"
#include "World.h"
#include "OpenGLWindow.h"
#include "CriticalSection.h"
#include "ThreadManagerSerial.h"
#include <SDL.h>

const int WIDTH = 800;
const int HEIGHT = 600;

World world;
OpenGLWindow window;

int *localData;

#ifdef USE_THREADS
#define THREAD_SPAWN_COUNT 2
#else
#define THREAD_SPAWN_COUNT 0
#endif

int threadRenderID = 0;
int threadUpdateID = 0;

int generalEntryPoint(void);

BOOL init(void);
void shutdown(void);

// OpenGLWindow callbacks
BOOL windowLoop(void);
void keypress(SDL_KeyboardEvent *event);

// ThreadManager callbacks
void worldRender();
void finalizeRender();
void worldUpdate();

int generalEntryPoint(void)
{
    BOOL okay = init();
    if(okay)
    {
        window.runWindowLoop(
            "Threaded Cross-Platform Game Development",
            WIDTH,
            HEIGHT,
            windowLoop,
            keypress
        );
        
        shutdown();
    }
    else
    {
        LOG("Initialization error, exiting...\n");
    }
    return 0;
}

#ifdef WIN32
// Windows entry point
// We don't use WinMain because SDL does some magic behind the scenes which invokes
// this main function from a predefined WinMain which does SDL-specific initialization
int main(int argc, char **argv) 
{
    int error = generalEntryPoint();
    
    return error;
}
#elif LINUX
// Linux entry point
int main(int argc, char **argv) 
{
    // imperative that multi-threaded Xlib programs call XInitThreads before any other X function
    XInitThreads();
    // there -- are you satisfied?

    int error = generalEntryPoint();
    
    return error;
}
#endif

BOOL init(void)
{
    // ensure SDL events are enabled
    SDL_Init(SDL_INIT_VIDEO);

    // make local data for main thread and any more we spawn
    localData = new int[THREAD_SPAWN_COUNT + 1];
    for(int i = 0; i <= THREAD_SPAWN_COUNT; i++)
    {
        localData[i] = 0;
    }

    // get a pointer to our ThreadManager
    ThreadManagerSerial &tm = ThreadManagerSerial::getThreadManagerSerial();
    BOOL okay = tm.init(THREAD_SPAWN_COUNT, &localData[0]);
    if(!okay)
    {
        LOG("Failed to initialize ThreadManager.\n");
        return false;
    }

    #ifdef USE_THREADS
    // setup update thread
    threadUpdateID = tm.assignThreadToWorkFunction(
        worldUpdate,
        NULL,
        &localData[1]
    );
    if(threadUpdateID == ThreadManager::INVALID_THREAD_ID)
    {
        LOG("Failed to assign thread for update task.\n");
        return false;
    }
    
    // setup rendering thread
    threadRenderID = tm.assignThreadToWorkFunction(
        worldRender,
        finalizeRender,
        &localData[2]
    );
    if(threadRenderID == ThreadManager::INVALID_THREAD_ID)
    {
        LOG("Failed to assign thread for render task.\n");
        return false;
    }
    
    // initialize the world, while informing it about the threads and the window
    world.init(threadUpdateID, threadRenderID, &window, WIDTH, HEIGHT);

    // start our threads
    
    // NOTE: uncomment these lines to begin in serial execution mode
    //tm.setThreadSerialized(threadUpdateID, true);
    //tm.setThreadSerialized(threadRenderID, true);
    
    tm.startThread(threadUpdateID);
    tm.startThread(threadRenderID);
    
    world.setPaused(false);
    
    #else
    world.init(0, 0, &window, WIDTH, HEIGHT);
    window.init();
    #endif
    
    return true;
}

void shutdown(void)
{
    SDL_Quit();

    // get a pointer to our ThreadManager
    ThreadManager &tm = ThreadManager::getThreadManager();
    
    #ifdef USE_THREADS
    // cleanup our threads
    tm.unassignThread(threadUpdateID);
    tm.unassignThread(threadRenderID);
    #endif
    
    tm.shutdown();
    
    world.shutdown();
    
    delete [] localData;
}

// OpenGLWindow callbacks
BOOL windowLoop(void)
{
    // get a pointer to our ThreadManager
    ThreadManagerSerial &tm = ThreadManagerSerial::getThreadManagerSerial();
    
    // ping the ThreadManager for serial work
    tm.doSerialWork();
    
    #ifndef USE_THREADS
    // if NOT using threads, call our functions directly
    worldUpdate();
    worldRender();
    #endif
    
    // ask the world if it's time to quit
    return world.isTimeToQuit();
}

void keypress(SDL_KeyboardEvent *event)
{
    world.handleKeypress(event);
}

// ThreadManager callbacks
void worldRender()
{
    // don't render if window is no good
    int currentWindowID = window.getWindowID();
    if(currentWindowID != 0)
    {
        BOOL goAheadAndRender = true;
    
        // is our window still valid?
        int *pLocalData = (int *)(ThreadManager::getThreadManager().getLocalData());
        if(*pLocalData != currentWindowID)
        {
            // need to reset the rendering context
            world.shutdownOpenGL();
            
            if(window.enableRender())
            {
                world.initOpenGL();
                
                *pLocalData = currentWindowID;
                
                // unpause the simulation
                world.setPaused(false);
            }
            else
            {
                goAheadAndRender = false;
            }
        }
    
        if(goAheadAndRender)
        {
            world.render();
            window.update();
        }
    }
}

void finalizeRender()
{
    window.disableRender();
}

void worldUpdate()
{
    world.update();
}
