// ------------------------------------------------------------------------------------
//
// Copyright  2006 Intel Corporation
// All Rights Reserved
// 
// The sample source code contained or described herein and all documents
// related to the sample source code ("Material") are owned by Intel Corporation
// or its suppliers or licensors. Title to the Material remains with Intel Corporation
// or its suppliers and licensors. The Material may contain trade secrets and
// proprietary and confidential information of Intel Corporation and its suppliers
// and licensors, and is protected by worldwide copyright and trade secret laws and
// treaty provisions. The sample source code is provided AS IS, with no warranties
// of any kind, express or implied. Except as expressly permitted by the Software
// license, neither Intel Corporation nor its suppliers assumes any responsibility
// or liability for any errors or inaccuracies that may appear herein.
//
// ------------------------------------------------------------------------------------
#include "stdafx.h"
#include "OpenGLWindow.h"
#include <SDL.h>

/** \brief constructor

Public constructor for OpenGLWindow.
*/
OpenGLWindow::OpenGLWindow(void)
{
    m_width = 0;
    m_height = 0;
    m_windowID = 0;
    m_lastWindowID = 0;
    m_fullscreen = false;
    m_respawnWindow = false;
}

/** \brief destructor

Public destructor for OpenGLWindow.
*/
OpenGLWindow::~OpenGLWindow(void)
{
}

/** \brief runs a simple event loop with callbacks

Call runWindowLoop() from your event thread to run a simple event loop.  The event loop will terminate
when the window is closed, or when the loopFunc function returns true.
\param windowName the title to be used for the window
\param width width of the window
\param height height of the window
\param loopFunc (optional) called repeatedly by the event loop, return true when it's time to quit
\param keypressFunc (optional) called whenever an SDL_KEYDOWN or SDL_KEYUP event occurs
*/
void OpenGLWindow::runWindowLoop(
    const char *windowName,
    int width,
    int height,
    CallbackLoopFunction loopFunc,
    CallbackKeypressFunction keypressFunc
)
{
    m_width = width;
    m_height = height;

    SDL_GL_SetAttribute(SDL_GL_DOUBLEBUFFER, 1);
    SDL_WM_SetCaption(windowName, NULL);
    
    m_respawnWindow = true;
    BOOL successfulRefresh = false;
    
    while(m_respawnWindow)
    {
        m_respawnWindow = false;
    
        m_windowID = ++m_lastWindowID;
        
        BOOL timeToQuit = false;
        SDL_Event event;
        while(!timeToQuit && !m_respawnWindow)
        {
            if(SDL_PollEvent(&event))
            {
                switch(event.type)
                {
                    case SDL_QUIT:
                        timeToQuit = true;
                        break;
                    
                    case SDL_KEYDOWN:
                        // intentional fall-through to next case
                    case SDL_KEYUP:
                        if(keypressFunc != NULL)
                        {
                            keypressFunc((SDL_KeyboardEvent *)&event);
                        }
                    
                    default:
                        break;
                }
            }
        
            if(!timeToQuit)
            {
                // call our loop function
                if(loopFunc != NULL)
                {
                    timeToQuit = loopFunc();
                }
            }
        }
        
        m_windowID = 0;
    }
}

/** \brief returns a unique ID number for the current window and OpenGL rendering context

Your rendering thread should call this function before rendering and compare the value returned against the
previous value.  Whenever the value returned by this function changes, the OpenGL rendering context has been
destroyed and your rendering thread must call enableRender() before making any more OpenGL calls.
Your rendering thread will also have to recreate the OpenGL state after the call to enableRender().
\return an integer representing a unique window ID for the current window and rendering context
\sa enableRender()
*/
int OpenGLWindow::getWindowID(void)
{
    return m_windowID;
}

/** \brief returns true if the window is fullscreen

Call this function to determine the fullscreen state of the current window.
\return true if the current window is fullscreen, false otherwise
\sa setFullscreen()
*/
BOOL OpenGLWindow::isFullscreen(void)
{
    return m_fullscreen;
}

/** \brief change the fullscreen state of the window

Call this function to change the fullscreen state of the current window.
\param fullscreen true if window should be make fullscreen, false otherwise
*/
void OpenGLWindow::setFullscreen(BOOL fullscreen)
{
    if(m_fullscreen != fullscreen)
    {
        m_fullscreen = fullscreen;
        
        // need to recreate the window
        m_respawnWindow = true;
    }
}
    
/** \brief prepare a thread to render to the window

Your rendering thread should call this function whenever the value returned by getWindowID() changes.
Your rendering thread will also have to recreate the OpenGL state after calling this function.
\return true if successful, false otherwise
\sa getWindowID(), disableRender()
*/
BOOL OpenGLWindow::enableRender(void)
{
    BOOL output = false;

    int flags = SDL_OPENGL;
    
    if(m_fullscreen)
    {
        flags |= SDL_FULLSCREEN;
    }
    
    SDL_Surface *surface = SDL_SetVideoMode(m_width, m_height, 0, flags);
    output = (surface != NULL);
        
    return output; 
}

/** \brief cleanup the OpenGL context when finished rendering

Your rendering thread should call this function when it relinquishes rendering duties.  This function
ensures that rendering can be successfully initiated by another thread.
\sa enableRender()
*/
void OpenGLWindow::disableRender(void)
{
    #ifdef LINUX
        // NOTE: this is an unfortunate hack to ensure that the OpenGL context attached to the thread
        // is abandoned, and a new window will be created to ensure enableRender is called
        // from the new thread
        SDL_SetVideoMode(m_width, m_height, 0, 0);
        m_respawnWindow = true;
    #endif
}

/** \brief update the window with your OpenGL calls

Your rendering thread should call this function after making all the OpenGL calls needed to draw
the current frame.
*/
void OpenGLWindow::update(void)
{ 
    if(m_windowID != 0)
    {
        SDL_GL_SwapBuffers();
    }
}
