// ------------------------------------------------------------------------------------
//
// Copyright  2006 Intel Corporation
// All Rights Reserved
// 
// The sample source code contained or described herein and all documents
// related to the sample source code ("Material") are owned by Intel Corporation
// or its suppliers or licensors. Title to the Material remains with Intel Corporation
// or its suppliers and licensors. The Material may contain trade secrets and
// proprietary and confidential information of Intel Corporation and its suppliers
// and licensors, and is protected by worldwide copyright and trade secret laws and
// treaty provisions. The sample source code is provided AS IS, with no warranties
// of any kind, express or implied. Except as expressly permitted by the Software
// license, neither Intel Corporation nor its suppliers assumes any responsibility
// or liability for any errors or inaccuracies that may appear herein.
//
// ------------------------------------------------------------------------------------
#include "stdafx.h"
#include "ThreadManager.h"
#include "CriticalSection.h"

const int THREAD_SLEEP_DURATION = 50;  //< milliseconds of sleep between status checks

#ifdef WIN32
    const ThreadManager::LocalDataKeyType ThreadManager::sm_invalidLocalDataKey = TLS_OUT_OF_INDEXES;
    const ThreadManager::ThreadType ThreadManager::sm_invalidThread = NULL;
#elif LINUX
    const ThreadManager::LocalDataKeyType ThreadManager::sm_invalidLocalDataKey = 0;
    const ThreadManager::ThreadType ThreadManager::sm_invalidThread = 0;
#endif

CriticalSection ThreadManager::sm_singletonCS;
ThreadManager *ThreadManager::sm_threadManager = NULL;

/** \brief static function for singleton access

Call getThreadManager() whenever you need access to the singleton ThreadManager.
*/
ThreadManager &ThreadManager::getThreadManager(void)
{
    if(sm_threadManager == NULL)
    {
        sm_singletonCS.enter();
        if(sm_threadManager == NULL)
        {
            sm_threadManager = new ThreadManager();
        }
        sm_singletonCS.leave();
    }
    return *sm_threadManager;
}   

/** \brief causes the current thread to yield control to a blocked thread, if one is waiting

Call yield() to cause the current thread to offer control to any waiting threads on the same processor.
If no threads are waiting, control will continue with the current thread.
*/
void ThreadManager::yield(void)
{
    #ifdef WIN32
        Sleep(0);
    #elif LINUX
        sched_yield();
    #endif
}

/** \brief causes the current thread to block for the specified number of milliseconds

Call sleep() to cause the current thread to block for a period of time.  When blocked, other threads
waiting on the same processor will be allowed to take control.
\param milliseconds number of milliseconds to block the current thread
*/
void ThreadManager::sleep(int milliseconds)
{
    ASSERT(milliseconds > 0);

    #ifdef WIN32
        Sleep(milliseconds);
    #elif LINUX
        usleep(milliseconds * 1000);
    #endif
}

/** \brief initialization function

Call init() once at the beginning of your program.  This function allocates the thread pool and creates synchronization
objects which are used by all other ThreadManager calls.  Whichever thread calls this function will be considered
the "primary thread" by the ThreadManager object, and will be the only thread allowed to call non-thread-safe functions.
\param threadPoolSize the number of threads to be created in the pool
\param primaryThreadLocalData (optional) pointer returned by getLocalData() when called from primary thread
\return true if initialization was successful, false otherwise
\sa getLocalData()
*/
BOOL ThreadManager::init(int threadPoolSize, void *primaryThreadLocalData)
{
    BOOL output = false;
    
    // make sure init isn't called on a succesfully initialized ThreadManager
    m_internalDataCS.enter();
    
    if(!m_initialized)
    {
        BOOL successful = true;
        
        // allocate a new thread pool
        if(threadPoolSize > 0)
        {
            // create the thread pool
            m_threadPool = new threadInfo[threadPoolSize];
            if(m_threadPool == NULL)
            {
                successful = false;
            }
            else
            {
                m_threadPoolSize = threadPoolSize;
            
                for(int i = 0; i < threadPoolSize; i++)
                {
                    // populate the threadInfo structure
                    m_threadPool[i].id = i;
                    m_threadPool[i].status = THREAD_STATUS_UNASSIGNED;
                    m_threadPool[i].workFunc = NULL;
                    m_threadPool[i].calls = 0;
                    m_threadPool[i].thread = sm_invalidThread;
                    m_threadPool[i].criticalSection = NULL;
                    m_threadPool[i].localData = NULL;
                    
                    // initialize the critical section
                    m_threadPool[i].criticalSection = new CriticalSection();
                    if(
                        (m_threadPool[i].criticalSection == NULL) ||
                        (m_threadPool[i].criticalSection->isValid() == false)
                    )
                    {
                        // remember this thread pool size so we can deallocate threads properly later
                        m_threadPoolSize = i;
                        break;
                    }
                    
                    // spawn the thread
                    #ifdef WIN32
                        // we use _beginthreadex to spawn threads because CreateThread and _beginthread
                        // have inconsistencies in corner cases
                        m_threadPool[i].thread = (ThreadType)_beginthreadex(
                            NULL,
                            0,
                            ThreadManager::threadCallback,
                            &(m_threadPool[i].id),
                            0,
                            NULL
                        );
                        if(m_threadPool[i].thread == sm_invalidThread)
                        {
                            successful = false;
                            // remember this thread pool size so we can deallocate threads properly later
                            m_threadPoolSize = i;
                            break;
                        }
                    #elif LINUX
                        if(pthread_create(
                           &(m_threadPool[i].thread),
                           NULL,
                           ThreadManager::threadCallback,
                           &(m_threadPool[i].id)
                        ) != 0)
                        {
                            successful = false;
                            // remember this thread pool size so we can deallocate threads properly later
                            m_threadPoolSize = i;
                            break;
                        }
                    #endif
                }
            }
        }
        
        if(successful)
        {
            // create the local data key and seed it with our primaryThreadLocalData
            #ifdef WIN32
                m_localDataKey = TlsAlloc();
                if(m_localDataKey == sm_invalidLocalDataKey)
                {
                    successful = false;
                }
                else
                {
                    // okay, instead of storing primaryThreadLocalData, we store a pointer to this
                    // which will let us know when we are executing from the primary thread
                    successful = TlsSetValue(m_localDataKey, this);
                }
            #elif LINUX
                if(pthread_key_create(&m_localDataKey, NULL) != 0)
                {
                    successful = false;
                }
                else
                {
                    // okay, instead of storing primaryThreadLocalData, we store a pointer to this
                    // which will let us know when we are executing from the primary thread
                    if(pthread_setspecific(m_localDataKey, this) != 0)
                    {
                        successful = false;
                    }
                }
            #endif
            
            // store the primaryThreadLocalData for later
            m_primaryThreadLocalData = primaryThreadLocalData;  
        }
        
        if(!successful)
        {
            // deallocate everything we made
            // don't wait for threads to terminate -- they were never assigned work functions
            cleanup(false);
        }
        else
        {        
            // everything worked, make sure we aren't initialized again
            m_initialized = true;   
            output = true;
        }
    }
    
    m_internalDataCS.leave();
    return output;
}

/** \brief shutdown function

Call shutdown() once from the primary thread at the end of your program.
*/
void ThreadManager::shutdown()
{
    m_internalDataCS.enter();

    if(m_initialized)
    {
        ASSERT(isPrimaryThread());
        // call cleanup and ensure that threads get terminated before returning
        cleanup(true);
        m_initialized = false;
    }
    
    m_internalDataCS.leave();
}

/** \brief assigns a thread in the thread pool to execute a function repeatedly

Call assignThreadToWorkFunction() from the primary thread to attempt to assign a thread
in the thread pool to a work function.
If a thread in the pool is available, it will be assigned to the work function, stopped,
and the id number will be returned.  Once the thread is started, the work function will be called
repeatedly by the thread.
The finalizeFunc will be called at the point when the thread will no longer call the work function.
This function can be used to do thread-specific cleanup.
\param workFunc function which will be called repeatedly by the assigned thread
\param finalizeFunc (optional) function which will be called when the thread will no longer call the work function
\param localData (optional) pointer returned by getLocalData() when called from assigned thread
\return ID of thread which will execute this task, or INVALID_THREAD_ID if no thread is available
\sa unassignThread(), getLocalData()
*/
int ThreadManager::assignThreadToWorkFunction(WorkFunction workFunc, FinalizeFunction finalizeFunc, void *localData)
{
    ASSERT(m_initialized);
    ASSERT(isPrimaryThread());
    
    int id = INVALID_THREAD_ID;
    
    // enter major critical section
    // we can now access thread pool data directly
    m_internalDataCS.enter();
    
    // find the next available thread id
    for(int i = 0; i < m_threadPoolSize; i++)
    {
        if(m_threadPool[i].status == THREAD_STATUS_UNASSIGNED)
        {
            id = i;
            break;
        }
    }
    
    if(id != INVALID_THREAD_ID)
    {
        m_threadPool[id].status = THREAD_STATUS_STOPPED;
        
        // populate the threadInfo record
        m_threadPool[id].workFunc = workFunc;
        m_threadPool[id].finalizeFunc = finalizeFunc;
        m_threadPool[id].calls = 0;
        m_threadPool[id].localData = localData;
    }
    
    m_internalDataCS.leave();
    
    return id;
}

/** \brief unassigns the thread with the given ID

Call unassignThread() from the primary thread to unassign the thread with the given ID.
The thread is not guaranteed to be immediately available for reassignment.
\param id ID of the thread to unassign 
\return true if the thread was unassigned, false otherwise
\sa assignThreadToWorkFunction()
*/
BOOL ThreadManager::unassignThread(int id)
{
    ASSERT(isPrimaryThread());
    
    // invalidate the thread
    // don't unassign already unassigned threads
    int status = testAndSetThreadStatus(
        id,
        ~THREAD_STATUS_UNASSIGNED,
        THREAD_STATUS_UNASSIGN_REQUEST
    );
    return (
        (status == THREAD_STATUS_UNASSIGN_REQUEST) ||
        (status == THREAD_STATUS_UNASSIGNED)
    );
}

/** \brief starts the thread, if possible

Call startThread() from the primary thread to cause the thread with the given ID
to begin executing its work function.
The work function will be called repeatedly, as often as possible.  Work functions can call
yield() or sleep() to defer their next execution, if desired.
\param id ID of the thread to start 
\return true if the thread was started, false otherwise
*/
BOOL ThreadManager::startThread(int id)
{
    ASSERT(isPrimaryThread());

    // make it go, if possible
    int status = testAndSetThreadStatus(
        id,
        THREAD_STATUS_STOPPED,
        THREAD_STATUS_RUNNING_REQUEST
    );
    return (
        (status == THREAD_STATUS_RUNNING_REQUEST) ||
        (status == THREAD_STATUS_RUNNING)
    );
}

/** \brief stops the thread, if possible

Call stopThread() from the primary thread to cause the thread with the given ID
to quit executing its work function.
\param id ID of the thread to stop 
\return true if the thread was stopped, false otherwise
*/
BOOL ThreadManager::stopThread(int id)
{
    ASSERT(isPrimaryThread());

    // make it stop, if possible
    int status = testAndSetThreadStatus(
        id,
        THREAD_STATUS_RUNNING | THREAD_STATUS_RUNNING_REQUEST,
        THREAD_STATUS_STOPPED
    );
    return (status == THREAD_STATUS_STOPPED);
}

/** \brief returns size of thread pool specified in init()

Call getThreadPoolSize() to get the number of threads in the thread pool.
\return size of the thread pool
\sa init()
*/
int ThreadManager::getThreadPoolSize(void) const
{
    return m_threadPoolSize;
}

/** \brief returns the local data associated with the current thread

This function returns a pointer to local data that was associated with the current thread
through a call to assignThreadToWorkFunction() or to init() for the primary thread.
\return local data pointer associated with the current thread
\sa assignThreadToWorkFunction(), init()
*/
void *ThreadManager::getLocalData(void) const
{
    void *output = getLocalDataRaw();
    if(output == this)
    {
        output = m_primaryThreadLocalData;
    }
    
    return output;
}

/** \brief returns true if the current thread is the primary thread

This function returns true if the current thread is the primary thread.  The primary thread
is defined as the thread which called the init() function.
\return true if called from primary thread, false otherwise
\sa init()
*/
BOOL ThreadManager::isPrimaryThread(void) const
{
    return(getLocalDataRaw() == this);
}

/** \brief retrieves the number of calls made by the thread with the given ID

Call getAndResetThreadCalls() to retrieve the number of calls a thread has made to its work function.
Calling this function will reset the call count to zero.
\param id ID of the thread to check
\return the number of calls the thread has made to its work function since the last call to getAndResetThreadCalls()
*/
int ThreadManager::getAndResetThreadCalls(int id)
{
    ASSERT(m_initialized);
    ASSERT((id >= 0) && (id < m_threadPoolSize));

    int output = 0;
    
    lockThread(id);
    output = m_threadPool[id].calls;
    m_threadPool[id].calls = 0;
    unlockThread(id);
    
    return output;
}

/** \brief protected constructor

This is the protected constructor for ThreadManager.
User code that needs a ThreadManager should call the static getThreadManager() function
\sa getThreadManager()
*/
ThreadManager::ThreadManager(void)
{
    m_initialized = false;
    m_threadPool = NULL;
    m_threadPoolSize = 0;
    m_primaryThreadLocalData = NULL;
    m_localDataKey = sm_invalidLocalDataKey;
}

/** \brief protected destructor

This is the protected destructor for ThreadManager.
*/
ThreadManager::~ThreadManager(void)
{
}

/** \brief protected function to clean up ThreadManager resources

This function cleans up ThreadManager resources during a shutdown() or a failed init().
Subclasses can override this function to clean up additional resources.
\sa shutdown(), init()
*/
void ThreadManager::cleanup(BOOL waitForThreadTermination)
{
    // we don't enter the internalDataCS, because this function is only called from init and shutdown
    // where the critical section is already entered

    // signal termination to all the threads in the pool
    for(int i = 0; i < m_threadPoolSize; i++)
    {
        // signal termination to thread
        testAndSetThreadStatus(i, THREAD_STATUS_ALL, THREAD_STATUS_TERMINATE_REQUEST);
    }
    
    // wait on all threads in turn, to terminate
    #ifdef LINUX
        // when we join the threads, only wait one second for the join
        timespec joinDelay;
        joinDelay.tv_sec = 1;  // wait only one second
    #endif
    
    for(int i = 0; i < m_threadPoolSize; i++)
    {
        // wait until this thread is done, then destroy it
        #ifdef WIN32
            // NOTE: we don't use WaitForMultipleObjects here, in order to make the cross-platform
            // code have more parity.  If we wanted to do so, we would move this code outside the loop,
            // allocate an array of thread objects, populate it from the thread pool, and destroy it
            // when finished.  We would still need to close each handle individually within the loop.
            
            if(waitForThreadTermination)
            {
                // wait forever
                WaitForSingleObject(m_threadPool[i].thread, INFINITE);
            }
            else
            {
                // wait at most one second
                WaitForSingleObject(m_threadPool[i].thread, 1000);
            }
            CloseHandle(m_threadPool[i].thread);
        #elif LINUX
            if(waitForThreadTermination)
            {
                // wait forever
                pthread_join(m_threadPool[i].thread, NULL);
            }
            else
            {
                // using options defined earlier, wait at most one second
                pthread_timedjoin_np(m_threadPool[i].thread, NULL, &joinDelay);
            }
            
            // whether or not we joined, detach the resources
            pthread_detach(m_threadPool[i].thread);
        #endif
        
        // also cleanup the critical section
        delete m_threadPool[i].criticalSection;
        m_threadPool[i].criticalSection = NULL;
    }

    delete [] m_threadPool;
    m_threadPool = NULL;
    m_threadPoolSize = 0;
    
    // destroy the local data key
    if(m_localDataKey != sm_invalidLocalDataKey)
    {
        #ifdef WIN32
            TlsFree(m_localDataKey);
        #elif LINUX
            pthread_key_delete(m_localDataKey);
        #endif
    }
    m_localDataKey = sm_invalidLocalDataKey;
}

/** \brief protected function that grants exclusive access to a thread's state

ThreadManager calls lockThread() before checking or modifying state information
of threads in the thread pool.
\param id ID of the thread whose state will be checked or modified
*/
void ThreadManager::lockThread(int id)
{
    // potentially called before initialized is true
    // so we don't assert on it
    ASSERT((id >= 0) && (id < m_threadPoolSize));
    ASSERT(m_threadPool[id].criticalSection != NULL);
    
    m_threadPool[id].criticalSection->enter();
}

/** \brief protected function that relinquishes exclusive access to a thread's state

ThreadManager calls unlockThread() when finished checking or modifying state information
of threads in the thread pool.
\param id ID of the thread whose state is no longer needed for checking or modification
*/
void ThreadManager::unlockThread(int id)
{
    // potentially called before initialized is true
    // so we don't assert on it
    ASSERT((id >= 0) && (id < m_threadPoolSize));
    ASSERT(m_threadPool[id].criticalSection != NULL);
    
    m_threadPool[id].criticalSection->leave();
}

/** \brief protected function that returns the state of the thread with the given ID

This function is used by the thread pool callback to process threads based upon their current state.
\param id ID of the thread whose state will be returned
\return the state of the thread
*/
int ThreadManager::getThreadStatus(int id) const
{
    // potentially, this function is called before initialized is true
    // so we don't assert on it
    ASSERT((id >= 0) && (id < m_threadPoolSize));
    
    return m_threadPool[id].status;
}

/** \brief protected function that atomically tests and changes thread state

This function is used throughout the ThreadManager to adjust thread state.
If the thread with the given ID is in a state that matches the condition mask,
the thread state will be changed to the given value.
\param id ID of the thread whose state will be tested and changed
\param condition a mask of thread states
\param value a thread state
\return value if the thread state was changed, otherwise the original thread state
*/
int ThreadManager::testAndSetThreadStatus(int id, int condition, int value)
{
    // potentially, this function is called before initialized is true
    // so we don't assert on it
    ASSERT((id >= 0) && (id < m_threadPoolSize));
    
    // atomic operation to test thread status for a mask of states
    // if status matches mask, set status to value
    lockThread(id);
    
    int output = m_threadPool[id].status;
    if((output & condition) != 0)
    {
        m_threadPool[id].status = value;
        output = value;
    }
 
    unlockThread(id); 
    return output;
}

/** \brief protected function that associates local data pointer with the thread with the given ID

This function is used by the thread pool callback to initialize a thread on its first run.
\param id ID of the thread which will have its local data pointer associated with the current thread
*/
void ThreadManager::associateLocalDataWithThread(int id)
{
    ASSERT(m_initialized);
    ASSERT((id >= 0) && (id < m_threadPoolSize));
    
    #ifdef WIN32
        TlsSetValue(m_localDataKey, m_threadPool[id].localData);
    #elif LINUX
        pthread_setspecific(m_localDataKey, m_threadPool[id].localData);
    #endif
}

/** \brief protected function that retrieves a pointer to local data associated with current thread

ThreadManager calls this function internally to get unprocessed local data associated with the current thread.
User code should call getLocalData().
\sa getLocalData()
*/
void *ThreadManager::getLocalDataRaw(void) const
{
    ASSERT(m_initialized);
    
    void *output = NULL;

    #ifdef WIN32
        output = TlsGetValue(m_localDataKey);
    #elif LINUX
        output = pthread_getspecific(m_localDataKey);
    #endif
    
    return output;
}

/** \brief protected function that calls the work function of the thread with the given ID,
and updates the call count

This function is used by the thread pool callback to invoke work functions.
\param id ID of the thread whose work function will be called
*/
void ThreadManager::callWorkFunctionForThread(int id)
{
    ASSERT(m_initialized);
    ASSERT((id >= 0) && (id < m_threadPoolSize));
    
    // no need to lock thread here since workFunc is guaranteed to be valid
    // for the duration of the function
    WorkFunction func = m_threadPool[id].workFunc;
    if(func != NULL)
    {
        (*func)();
        
        #ifdef WIN32
            // Windows has a lightweight way to do an atomic increment
            InterlockedIncrement(&m_threadPool[id].calls);
        #else
            lockThread(id);
            m_threadPool[id].calls++;
            unlockThread(id);
        #endif
    }
}

/** \brief protected function that calls the finalize function of the thread with the given ID

This function is called by the ThreadManager class whenever a thread is unassigned.
\param id ID of the thread whose finalize function will be called
*/
void ThreadManager::callFinalizeFunctionForThread(int id)
{
    ASSERT(m_initialized);
    ASSERT((id >= 0) && (id < m_threadPoolSize));
    
    // no need to lock thread here since finalizeFunc is guaranteed to be valid
    // for the duration of the function
    FinalizeFunction func = m_threadPool[id].finalizeFunc;
    if(
        (func != NULL) &&
        (m_threadPool[id].calls > 0)
    )
    {
        (*func)();
    }
}

/** \brief protected function that processes threads in thread pool

This function is used by the thread pool callback to process the thread based upon its current state.
Subclasses can override this function to handle new thread states.
\param id ID of the thread to handle
\param status current state of the thread
*/
void ThreadManager::callbackHandleStatus(int id, int status)
{
    int statusResult = THREAD_STATUS_UNASSIGNED;

    switch(status)
    {
        case THREAD_STATUS_RUNNING_REQUEST:
            statusResult = testAndSetThreadStatus(id, THREAD_STATUS_RUNNING_REQUEST, THREAD_STATUS_RUNNING);
            if(statusResult != THREAD_STATUS_RUNNING)
            {
                // sleep it off and try again later
                sleep(THREAD_SLEEP_DURATION);
                break;
            }
            // it worked, so prepare for running
            associateLocalDataWithThread(id);
            // intentional fall-through
        case THREAD_STATUS_RUNNING:
            callWorkFunctionForThread(id);
            yield();
            break;
            
        case THREAD_STATUS_UNASSIGN_REQUEST:
            callFinalizeFunctionForThread(id);
            testAndSetThreadStatus(id, THREAD_STATUS_UNASSIGN_REQUEST, THREAD_STATUS_UNASSIGNED);
            // whether successful or not, fall through to default case
        default:
            // sleep for a bit
            sleep(THREAD_SLEEP_DURATION);
            break;
    }
}

/** \brief protected function that is called directly by threads in the thread pool

This is the thread pool callback function, called by each thread in the thread pool.
\param param a pointer to the ID of the thread which made the call.
*/
#ifdef WIN32
unsigned int __stdcall ThreadManager::threadCallback(void *param)
#elif LINUX
void *ThreadManager::threadCallback(void *param)
#endif
{
    ASSERT(param != NULL);

    // param is just the thread id number
    int id = *(int *)param;
    
    // get the ThreadManager
    ThreadManager &tm = ThreadManager::getThreadManager();
    
    int status = tm.getThreadStatus(id);
    while(status != THREAD_STATUS_TERMINATE_REQUEST)
    {
        // take whatever actions are appropriate for this status
        tm.callbackHandleStatus(id, status);
    
        // update status
        status = tm.getThreadStatus(id);
    }

    #ifdef WIN32
    _endthreadex(0);
    #endif
 
    return 0; 
}
