// ------------------------------------------------------------------------------------
//
// Copyright  2006 Intel Corporation
// All Rights Reserved
// 
// The sample source code contained or described herein and all documents
// related to the sample source code ("Material") are owned by Intel Corporation
// or its suppliers or licensors. Title to the Material remains with Intel Corporation
// or its suppliers and licensors. The Material may contain trade secrets and
// proprietary and confidential information of Intel Corporation and its suppliers
// and licensors, and is protected by worldwide copyright and trade secret laws and
// treaty provisions. The sample source code is provided AS IS, with no warranties
// of any kind, express or implied. Except as expressly permitted by the Software
// license, neither Intel Corporation nor its suppliers assumes any responsibility
// or liability for any errors or inaccuracies that may appear herein.
//
// ------------------------------------------------------------------------------------
#ifndef THREADMANAGER_H
#define THREADMANAGER_H

#include "stdafx.h"
#include "CriticalSection.h"

#ifdef WIN32
#include <process.h>
#elif LINUX
#include <pthread.h>
#include <unistd.h>
#endif

/** \brief creates and manages a thread pool

ThreadManager creates and manages a fixed-size thread pool.  Call assignThreadToWorkFunction() to dedicate
a thread in the pool to a task.  The supplied work function will be called repeatedly by the thread.
Work functions can be made to block, in order to avoid unnecessary calls.
*/
class ThreadManager
{
public:
    // type definition for work functions and finalize functions passed to createThread()
    typedef void (*WorkFunction)();
    typedef void (*FinalizeFunction)();
    
    static const int INVALID_THREAD_ID = -1;  //< thread ID number returned on error

    // thread-safe utility functions
    static ThreadManager &getThreadManager(void);
    
    static void yield(void);
    static void sleep(int milliseconds);
    
    // non thread-safe functions (call only from primary thread)
    BOOL init(int threadPoolSize, void *primaryThreadLocalData = NULL);
    void shutdown(void);
    int assignThreadToWorkFunction(WorkFunction workFunc, FinalizeFunction finalizeFunc = NULL, void *threadLocalData = NULL);
    BOOL unassignThread(int id);
    virtual BOOL startThread(int id);
    virtual BOOL stopThread(int id);
    
    // thread-safe functions
    int getThreadPoolSize(void) const;
    void *getLocalData(void) const;
    BOOL isPrimaryThread(void) const;
    virtual int getAndResetThreadCalls(int id);    
    
protected:
    // constants for thread status
    enum {
        THREAD_STATUS_ALL                       = ~0,
        THREAD_STATUS_UNASSIGNED                = 1 << 0,
        THREAD_STATUS_STOPPED                   = 1 << 1,
        THREAD_STATUS_RUNNING                   = 1 << 2,
        THREAD_STATUS_RUNNING_REQUEST           = 1 << 3,
        THREAD_STATUS_UNASSIGN_REQUEST          = 1 << 4,
        THREAD_STATUS_TERMINATE_REQUEST         = 1 << 5,
        THREAD_STATUS_MAX                       = 1 << 5
    };

    #ifdef WIN32
        // local data key
        typedef DWORD LocalDataKeyType;
        // thread object
        typedef HANDLE ThreadType;
    #elif LINUX
        // local data key
        typedef pthread_key_t LocalDataKeyType;
        // thread object
        typedef pthread_t ThreadType;
    #endif
    
    static const LocalDataKeyType sm_invalidLocalDataKey;
    static const ThreadType sm_invalidThread;
    static CriticalSection sm_singletonCS;
    static ThreadManager* sm_threadManager;
    
    ThreadManager(void);
    ~ThreadManager(void);

    void cleanup(BOOL waitForThreadTermination);
    
    void lockThread(int id);
    void unlockThread(int id);
    
    int getThreadStatus(int id) const;
    int testAndSetThreadStatus(int id, int condition, int value);
    
    void associateLocalDataWithThread(int id);
    void *getLocalDataRaw(void) const;
    
    void callWorkFunctionForThread(int id);
    void callFinalizeFunctionForThread(int id);
    virtual void callbackHandleStatus(int id, int status);
    
    #ifdef WIN32
        // thread pool callback
        static unsigned int __stdcall threadCallback(void *param);
    #elif LINUX
        // thread pool callback
        static void *threadCallback(void *param);
    #endif

    // threadInfo structure keeps track of each thread in the pool
    typedef struct {
        int id;
        int status;
        WorkFunction workFunc;
        FinalizeFunction finalizeFunc;
        long calls;  // long to take advantage of InterlockedIncrement on WIN32
        ThreadType thread;
        CriticalSection *criticalSection;
        void *localData;
    } threadInfo;
    
    BOOL m_initialized;
    
    // thread pool
    threadInfo *m_threadPool;
    int m_threadPoolSize;
    void *m_primaryThreadLocalData;
    LocalDataKeyType m_localDataKey;
    CriticalSection m_internalDataCS;
};

#endif
