// ------------------------------------------------------------------------------------
//
// Copyright  2006 Intel Corporation
// All Rights Reserved
// 
// The sample source code contained or described herein and all documents
// related to the sample source code ("Material") are owned by Intel Corporation
// or its suppliers or licensors. Title to the Material remains with Intel Corporation
// or its suppliers and licensors. The Material may contain trade secrets and
// proprietary and confidential information of Intel Corporation and its suppliers
// and licensors, and is protected by worldwide copyright and trade secret laws and
// treaty provisions. The sample source code is provided AS IS, with no warranties
// of any kind, express or implied. Except as expressly permitted by the Software
// license, neither Intel Corporation nor its suppliers assumes any responsibility
// or liability for any errors or inaccuracies that may appear herein.
//
// ------------------------------------------------------------------------------------
#include "stdafx.h"
#include "ThreadManagerSerial.h"
#include "CriticalSection.h"

/** \brief static function for singleton access

Call getThreadManagerSerial() whenever you need access to the singleton ThreadManagerSerial.  If your
program is using the singleton, and it mixes calls to ThreadManagerSerial and ThreadManager, be sure to
call this function before any call to ThreadManager::getThreadManager(), or the singleton will not be
instantiated properly.
*/
ThreadManagerSerial &ThreadManagerSerial::getThreadManagerSerial(void)
{
    if(sm_threadManager == NULL)
    {
        sm_singletonCS.enter();
        if(sm_threadManager == NULL)
        {
            sm_threadManager = new ThreadManagerSerial();
        }
        sm_singletonCS.leave();
    }
    
    return (ThreadManagerSerial &)*sm_threadManager;
}   

/** \brief call work functions for serialized threads

Call doSerialWork() in the event loop of your primary thread.  This function will call all work functions
that have been assigned to serialized threads, and then yield() the primary thread.  The order of execution of
work functions is not guaranteed.
\sa setThreadSerialized()
*/
void ThreadManagerSerial::doSerialWork()
{
    ASSERT(isPrimaryThread());

    // iterate over threadPool and find threads running in serial
    // call their work and finalize functions
    int status;
    for(int i = 0; i < m_threadPoolSize; i++)
    {
        status = getThreadStatus(i);
        switch(status)
        {
            case THREAD_STATUS_SERIAL_RUNNING:
                callWorkFunctionForThread(i);
                break;
        
            case THREAD_STATUS_SERIAL_TO_PARALLEL_STOPPED_REQUEST:
                callFinalizeFunctionForThread(i);
                testAndSetThreadStatus(i, THREAD_STATUS_SERIAL_TO_PARALLEL_STOPPED_REQUEST, THREAD_STATUS_STOPPED);
                break;
        
            case THREAD_STATUS_SERIAL_TO_PARALLEL_RUNNING_REQUEST:
                callFinalizeFunctionForThread(i);
                testAndSetThreadStatus(i, THREAD_STATUS_SERIAL_TO_PARALLEL_RUNNING_REQUEST, THREAD_STATUS_RUNNING_REQUEST);
                break;
        
            default:
                // not in a serial state
                break;
        }
    }
    
    // increment the serial call count
    #ifdef WIN32
        // Windows has a lightweight way to do an atomic increment
        InterlockedIncrement(&m_serialCalls);
    #else
        m_internalDataCS.enter();
        m_serialCalls++;
        m_internalDataCS.leave();
    #endif
    
    // yield to OS to prevent throttling
    yield();
}

/** \brief unassigns the thread with the given ID

Call unassignThread() from the primary thread to unassign the thread with the given ID.
The thread is not guaranteed to be immediately available for reassignment.
\param id ID of the thread to unassign 
\return true if the thread was unassigned, false otherwise
\sa assignThreadToWorkFunction()
*/
BOOL ThreadManagerSerial::unassignThread(int id)
{
    ASSERT(isPrimaryThread());
   
    BOOL output = false;
    
    // invalidate the thread
    // check for all states that are no longer handled by thread callback
    // these are states that are fundamentally serialized
    int status = testAndSetThreadStatus(
        id,
        (
            THREAD_STATUS_SERIAL_STOPPED |
            THREAD_STATUS_SERIAL_RUNNING |
            THREAD_STATUS_SERIAL_TO_PARALLEL_STOPPED_REQUEST |
            THREAD_STATUS_SERIAL_TO_PARALLEL_RUNNING_REQUEST
        ),
        THREAD_STATUS_UNASSIGNED
    );
    output = (status == THREAD_STATUS_UNASSIGNED);
    
    if(output)
    {
        // if we unassigned the thread in a serial way, call the finalize function
        callFinalizeFunctionForThread(id);
    }
    else
    {
        // must not be serial, try to unassign it in a parallel way
        output = ThreadManager::unassignThread(id);
    }
    
    return output; 
}

/** \brief starts the thread, if possible

Call startThread() from the primary thread to start executing the work function assigned to the thread
with the given ID, even if that thread has been serialized by a call to setThreadSerialized().
If the thread has been serialized, the work function will only be executed
if the thread it is assigned to has been started.
\warning This function will invalidate a pending serialization or parallelization request.  Check isThreadSerialized()
to ensure proper state before calling this function.
\param id ID of the thread to start 
\return true if the thread was started, false otherwise
\sa setThreadSerialized()
*/
BOOL ThreadManagerSerial::startThread(int id)
{
    ASSERT(isPrimaryThread());

    // make it go, if possible
    BOOL output = ThreadManager::startThread(id);
    if(!output)
    {
        // deal with extra states introduced to parallel execution
        // NOTE: this will invalidate a pending serialization request
        int status = testAndSetThreadStatus(
            id,
            THREAD_STATUS_PARALLEL_TO_SERIAL_STOPPED_REQUEST,
            THREAD_STATUS_RUNNING_REQUEST
        );
        output = output || (status == THREAD_STATUS_RUNNING_REQUEST);
    
        // also deal with serial states
        // NOTE: this will invalidate a pending parallelization request
        status = testAndSetThreadStatus(
            id,
            (
                THREAD_STATUS_SERIAL_STOPPED |
                THREAD_STATUS_SERIAL_TO_PARALLEL_STOPPED_REQUEST
            ),
            THREAD_STATUS_SERIAL_RUNNING
        );
        output = output || (status == THREAD_STATUS_SERIAL_RUNNING);
    }
    return output;
}

/** \brief stops the thread, if possible

Call stopThread() from the primary thread to stop executing the work function assigned to the thread
with the given ID, even if that thread has been serialized by a call to setThreadSerialized().
If the thread has been serialized, the work function will be executed
unless the thread it is assigned to is stopped.
\warning This function will invalidate a pending serialization or parallelization request.  Check isThreadSerialized()
to ensure proper state before calling this function.
\param id ID of the thread to stop 
\return true if the thread was stopped, false otherwise
\sa setThreadSerialized()
*/
BOOL ThreadManagerSerial::stopThread(int id)
{
    ASSERT(isPrimaryThread());

    // make it stop, if possible
    BOOL output = ThreadManager::stopThread(id);
    if(!output)
    {
        // deal with extra states introduced to parallel execution
        // NOTE: this will invalidate a pending serialization request
        int status = testAndSetThreadStatus(
            id,
            THREAD_STATUS_PARALLEL_TO_SERIAL_RUNNING_REQUEST,
            THREAD_STATUS_STOPPED
        );
        output = output || (status == THREAD_STATUS_STOPPED);
    
        // also deal with serial states
        // NOTE: this will invalidate a pending parallelization request
        status = testAndSetThreadStatus(
            id,
            (
                THREAD_STATUS_SERIAL_RUNNING |
                THREAD_STATUS_SERIAL_TO_PARALLEL_RUNNING_REQUEST
            ),
            THREAD_STATUS_SERIAL_STOPPED
        );
        output = output || (status == THREAD_STATUS_SERIAL_STOPPED);
    }
    return output;
}

/** \brief serialize or deserialize the thread with the given ID

Call setThreadSerialized() from the primary thread to change the serialization status of a thread.
Threads that are serialized will have their work functions executed during a call to doSerialWork().
When transitioning from serial to parallel, threads will call their finalize function before relinquishing
control of the work function.
\param id ID of the thread to change
\param serial true to serialize the thread, false to deserialize it
\return true if the change was successful, false otherwise
\sa isThreadSerialized(), doSerialWork()
*/
BOOL ThreadManagerSerial::setThreadSerialized(int id, BOOL serial)
{
    ASSERT(isPrimaryThread());

    BOOL output = false;
    int status;

    if(serial)
    {
        // make serial
        // don't know if it's running or not, so try both
        // but try stopped state first
        status = testAndSetThreadStatus(
            id,
            THREAD_STATUS_STOPPED,
            THREAD_STATUS_PARALLEL_TO_SERIAL_STOPPED_REQUEST
        );
        output = output || (status == THREAD_STATUS_PARALLEL_TO_SERIAL_STOPPED_REQUEST);
        
        status = testAndSetThreadStatus(
            id,
            THREAD_STATUS_RUNNING | THREAD_STATUS_RUNNING_REQUEST,
            THREAD_STATUS_PARALLEL_TO_SERIAL_RUNNING_REQUEST
        );
        output = output || (status == THREAD_STATUS_PARALLEL_TO_SERIAL_RUNNING_REQUEST);
    }
    else
    {
        // make parallel
        // don't know if it's running or not, so try both
        // but try stopped state first
        status = testAndSetThreadStatus(
            id,
            THREAD_STATUS_SERIAL_STOPPED,
            THREAD_STATUS_SERIAL_TO_PARALLEL_STOPPED_REQUEST
        );
        output = output || (status == THREAD_STATUS_SERIAL_TO_PARALLEL_STOPPED_REQUEST);
        
        status = testAndSetThreadStatus(
            id,
            THREAD_STATUS_SERIAL_RUNNING,
            THREAD_STATUS_SERIAL_TO_PARALLEL_RUNNING_REQUEST
        );
        output = output || (status == THREAD_STATUS_SERIAL_TO_PARALLEL_RUNNING_REQUEST);
    }
    
    return output;
}

/** \brief retrieves the number of calls made by the thread with the given ID

Call getAndResetThreadCalls() to retrieve the number of calls a thread has made to its work function.
Calling this function will reset the call count to zero.
Pass in an ID of INVALID_THREAD_ID to retrieve the number of calls made to doSerialWork()
from the primary thread.
\param id ID of the thread to check
\return the number of calls the thread has made to its work function since the last call to getAndResetThreadCalls()
\sa doSerialWork()
*/
int ThreadManagerSerial::getAndResetThreadCalls(int id)
{
    int output = 0;
    
    if(id == INVALID_THREAD_ID)
    {
        m_internalDataCS.enter();
        output = m_serialCalls;
        m_serialCalls = 0;
        m_internalDataCS.leave();
    }
    else
    {
        output = ThreadManager::getAndResetThreadCalls(id);
    }
    
    return output;
}

/** \brief get serialized status of the thread with the given ID

\param id ID of the thread to check 
\return true if the thread is serialized, false otherwise
\sa setThreadSerialized()
*/
BOOL ThreadManagerSerial::isThreadSerialized(int id)
{
    BOOL output = false;
    
    int status = getThreadStatus(id);
    if(
        (status == THREAD_STATUS_SERIAL_STOPPED) ||
        (status == THREAD_STATUS_SERIAL_RUNNING) ||
        (status == THREAD_STATUS_SERIAL_TO_PARALLEL_RUNNING_REQUEST) ||
        (status == THREAD_STATUS_SERIAL_TO_PARALLEL_STOPPED_REQUEST)
    )
    {
        output = true;
    }
    
    return output;
}

/** \brief protected constructor

This is the protected constructor for ThreadManagerSerial.
User code that needs a ThreadManagerSerial should call the static getThreadManagerSerial() function
\sa getThreadManagerSerial()
*/
ThreadManagerSerial::ThreadManagerSerial(void)
{
    m_serialCalls = 0;
}

/** \brief protected destructor

This is the protected destructor for ThreadManagerSerial.
*/
ThreadManagerSerial::~ThreadManagerSerial(void)
{
}

/** \brief protected function that processes threads in thread pool

This function is used by the thread pool callback to process the thread based upon its current state.
Subclasses can override this function to handle new thread states.  ThreadManagerSerial overrides
this function to handle serialized threads correctly.
\param id ID of the thread to handle
\param status current state of the thread
*/
void ThreadManagerSerial::callbackHandleStatus(int id, int status)
{
    switch(status)
    {
        case THREAD_STATUS_PARALLEL_TO_SERIAL_RUNNING_REQUEST:
            // change to run serially
            // call the finalize function
            callFinalizeFunctionForThread(id);
            testAndSetThreadStatus(id, THREAD_STATUS_PARALLEL_TO_SERIAL_RUNNING_REQUEST, THREAD_STATUS_SERIAL_RUNNING);
            break;
            
        case THREAD_STATUS_PARALLEL_TO_SERIAL_STOPPED_REQUEST:
            // change to run serially, but stopped
            // call the finalize function
            callFinalizeFunctionForThread(id);
            testAndSetThreadStatus(id, THREAD_STATUS_PARALLEL_TO_SERIAL_STOPPED_REQUEST, THREAD_STATUS_SERIAL_STOPPED);
            break;
            
        default:
            // fall back on ThreadManager's implementation
            ThreadManager::callbackHandleStatus(id, status);
            break;
    }
}
