// ------------------------------------------------------------------------------------
//
// Copyright  2006 Intel Corporation
// All Rights Reserved
// 
// The sample source code contained or described herein and all documents
// related to the sample source code ("Material") are owned by Intel Corporation
// or its suppliers or licensors. Title to the Material remains with Intel Corporation
// or its suppliers and licensors. The Material may contain trade secrets and
// proprietary and confidential information of Intel Corporation and its suppliers
// and licensors, and is protected by worldwide copyright and trade secret laws and
// treaty provisions. The sample source code is provided AS IS, with no warranties
// of any kind, express or implied. Except as expressly permitted by the Software
// license, neither Intel Corporation nor its suppliers assumes any responsibility
// or liability for any errors or inaccuracies that may appear herein.
//
// ------------------------------------------------------------------------------------
#include "stdafx.h"
#include "World.h"
#include "OpenGLWindow.h"
#include "CriticalSection.h"
#include "ThreadManagerSerial.h"

// GLF include
#include "../GLF/glf.h"

// GLF license:
/*
==============================================================================
|  GLF Library
|  Version 1.4
| 
|  Author: Roman Podobedov
|  Email: romka@ut.ee
|  WEB: http://romka.demonews.com
|  Release Date: 18 May 2001
|  
|  Copyright (C) 2000-2001, Romka Graphics
|  This library is freely distributable without any license or permissions
|  for non-commercial usage. You can use this library in any non-commercial
|  program. In each program, where You use this library You should keep 
|  this header (author name and coordinates)!
|  For commercial usage, please contact me: romka@ut.ee
==============================================================================
*/

// OpenGL constants
#define PI                      3.14159f
#define TWOPI                   2.0f * PI
#define FOV                     45.0f
#define FOV_RADIANS             ((FOV * PI) / 180.0f)
#define BACKGROUND_OFFSET_Z     3.0f
#define FOREGROUND_OFFSET_Z     2.0f
#define LABEL_OFFSET_Z          1.0f
#define NEAR_CLIP               0.1f
#define FAR_CLIP                10.0f

#define CPS_FREQUENCY_MILLISECONDS  2000

World::World(void)
{
}

World::~World(void)
{
}

void World::init(int threadUpdateID, int threadRenderID, OpenGLWindow *window, int width, int height)
{
    m_threadUpdateID = threadUpdateID;
    m_threadRenderID = threadRenderID;
    m_window = window;
    
    m_paused = false;
    setPaused(true);
    
    m_timeToQuit = false;

    m_width = width;
    m_height = height;
    
    m_aspect = 1.33f;
    if(m_height > 0)
    {
        m_aspect = (float)((float)m_width / (float)m_height);
    }

    // calculate how much of world is visible through window
    // extentXFactor can be multiplied by a Z value to determine width
    // of visible world coordinates at that Z
    m_extentYFactor = 2.0f * tan(FOV_RADIANS * 0.5f);
    m_extentXFactor = m_extentYFactor * m_aspect;
    if(m_extentXFactor < m_extentYFactor)
    {
        m_maxRadius = m_extentYFactor * 0.55f;
    }
    else
    {
        m_maxRadius = m_extentXFactor * 0.55f;
    }
    m_foregroundRadius = m_maxRadius * FOREGROUND_OFFSET_Z;
    m_blackHoleRadius = 0.025f * m_foregroundRadius;
    m_foregroundElementQuadricObj = NULL;
    m_foregroundElementRadius = radiusForMass(1);
    m_blackHoleMass = 500.0f;
    m_blackHoleState = BLACK_HOLE_STABLE;
    
    // seed the timing function, and the random function
    int millis = calculateTimeDeltaMilliseconds();
    m_millisecondsPassed = 0;
    srand(millis);

    // initialize visual elements
    m_backgroundCount = BACKGROUND_STEP;
    m_backgroundMax = 0;
    generateBackground();
    
    m_foregroundCount = FOREGROUND_STEP;
    m_foregroundMax = 0;
    generateForeground();
    
    m_cpsBuffer[0] = '\0';
    
    #ifdef WIN32
        SPRINTF(m_instructionBuffer, " ESC to quit, TAB to toggle serial/parallel, F to toggle fullscreen ");
    #elif LINUX
        SPRINTF(m_instructionBuffer, " ESC to quit, TAB to toggle serial/parallel ");
    #endif
    SPRINTF(m_instructionBuffer2, " Z & X to adjust render, . & / to adjust update ");
}

void World::shutdown(void)
{
}

void World::initOpenGL(void)
{
    glClearColor(0.0f, 0.0f, 0.0f, 1.0f);
    
    glPointSize(2.0f);
    
    // do GLF initialization
    glfInit();
	m_font = glfLoadBMFFont("fonts/arial.bmf");
	glfSetCurrentBMFFont(m_font);
	glfBitmapStringCentering(true);
	
	// make call lists
	int callList = glGenLists(1);
	
	m_foregroundElementCallList = callList;
	m_foregroundElementQuadricObj = gluNewQuadric();
	gluQuadricDrawStyle(m_foregroundElementQuadricObj, GLU_FILL);
	glNewList(m_foregroundElementCallList, GL_COMPILE);
	gluSphere(m_foregroundElementQuadricObj, 1.0, 8, 2);
	glEndList();
	
	// prepare the viewport and projection matrix
	glViewport(0, 0, m_width, m_height);

    glMatrixMode(GL_PROJECTION);
    glLoadIdentity();
    gluPerspective(FOV, m_aspect, NEAR_CLIP, FAR_CLIP);
    glMatrixMode(GL_MODELVIEW);
}

void World::shutdownOpenGL(void)
{
    if(m_foregroundElementQuadricObj != NULL)
    {
        gluDeleteQuadric(m_foregroundElementQuadricObj);
        m_foregroundElementQuadricObj = NULL;
    }
}

BOOL World::isPaused()
{
    return m_paused;
}

void World::setPaused(BOOL paused)
{
    m_paused = paused;
}
    
void World::update(void)
{
    if(m_paused)
    {
        return;
    }

    // figure out the time delta
    int milliseconds = calculateTimeDeltaMilliseconds();
    
    // do we need to update our calls per second info?
    m_millisecondsPassed += milliseconds;
    if(m_millisecondsPassed > CPS_FREQUENCY_MILLISECONDS)
    {
        // update number of calls for all our threads
        updateCallsPerSecond(m_millisecondsPassed);
        m_millisecondsPassed = 0;
    }
    
    float seconds = ((float)milliseconds / 1000.0f);
    
    // update the foreground elements
    ForegroundElement *e1 = NULL;
    ForegroundElement *e2 = NULL;
    float deltaX = 0.0f;
    float deltaY = 0.0f;
    float collisionRadius = 0.0f;
    float distanceSquared = 0.0f;
    float distance = 0.0f;
    float factor = 0.0f;
    float survivorMass = 0.0f;
    
    const float gravitationConstant = 0.000005f;
    const float blackHoleMaxMass = 10000.0f;
    
    // figure out what's going on with the black hole
    if(
        (m_blackHoleState == BLACK_HOLE_STABLE) &&
        (m_blackHoleMass > 600.0f)
    )
    {
        m_blackHoleState = BLACK_HOLE_EXPLODE; 
    }
    else if(
        (m_blackHoleState == BLACK_HOLE_EXPLODE) &&
        (m_blackHoleMass < 500.0f)
    )
    {
        m_blackHoleState = BLACK_HOLE_STABLE;
    }
    
    for(int i = 0; i < m_foregroundCount; ++i)
    {
        e1 = &m_foreground[i];
        
        if(e1->mass > 0.0f)
        {
            // figure out distance from center
            distanceSquared = ((e1->location[0] * e1->location[0]) + (e1->location[1] * e1->location[1]));
            distance = sqrt(distanceSquared);
            
            // if location is in the black hole, then reset
            // if location is out of the universe, then reset
            if(distance < m_blackHoleRadius)
            {
                m_blackHoleMass += e1->mass;
                if(m_blackHoleMass > blackHoleMaxMass)
                {
                    m_blackHoleMass = blackHoleMaxMass;
                }
                e1->mass = 0.0f;
            }
            else if(distance > m_foregroundRadius)
            {
                e1->mass = 0.0f;
            }
            else
            { 
                // apply attraction force to black hole
                factor = ((gravitationConstant * seconds) / (distance * distanceSquared)) * m_blackHoleMass;
                            
                e1->velocity[0] += (factor * -e1->location[0]);
                e1->velocity[1] += (factor * -e1->location[1]);
                
                // apply attraction to every other body
                for(int j = (i + 1); j < m_foregroundCount; ++j)
                {
                    e2 = &m_foreground[j];
                    if(e2->mass <= 0.0f)
                    {
                        continue;
                    }
                
                    // determine distance between bodies
                    deltaX = (e2->location[0] - e1->location[0]);
                    deltaY = (e2->location[1] - e1->location[1]);
                    distanceSquared = ((deltaX * deltaX) + (deltaY * deltaY));
                    collisionRadius = (e1->radius + e2->radius);
                    if(distanceSquared < ((collisionRadius * collisionRadius) * 0.5f))
                    {
                        // collision!
                        // the more massive one survives
                        // make sure survivor is pointed to by e1
                        if(e1->mass < e2->mass)
                        {
                            // want the more massive elements to have low indicies
                            // so swap e2 data with e1
                            swapForegroundElement(e1, e2);
                        }
                        
                        // figure out new velocity vector
                        // obey conservation of momentum
                        survivorMass = e1->mass + e2->mass;
                        
                        // should we model conservation of momentum?
                        if(e1->mass < e2->mass * 2.5f)
                        {
                            // masses are close, so obey conservation of momentum
                            e1->velocity[0] = (
                                ((e1->velocity[0] * e1->mass) + (e2->velocity[0] * e2->mass)) / survivorMass
                            );
                            e1->velocity[1] = (
                                ((e1->velocity[1] * e1->mass) + (e2->velocity[1] * e2->mass)) / survivorMass
                            );
                        }
                        
                        e1->mass = survivorMass;
                        e1->radius = radiusForMass(survivorMass);
                        e2->mass = 0.0f;
                    }
                    else
                    {
                        distance = sqrt(distanceSquared);
                        
                        factor = ((gravitationConstant * seconds) / (distance * distanceSquared));
                            
                        e1->velocity[0] += (factor * deltaX) * e2->mass;
                        e1->velocity[1] += (factor * deltaY) * e2->mass;
                        
                        e2->velocity[0] += (factor * -deltaX) * e1->mass;
                        e2->velocity[1] += (factor * -deltaY) * e1->mass;
                    }
                }
                
                if(e1->mass > 0.0f)
                {
                    // apply velocity
                    e1->location[0] += (e1->velocity[0] * seconds);
                    e1->location[1] += (e1->velocity[1] * seconds);
                    
                    // apply rotation
                    e1->rotation += (e1->rotationVelocity * seconds);
                    while(e1->rotation > TWOPI)
                    {
                        e1->rotation -= TWOPI;
                    }
                    while(e1->rotation < -TWOPI)
                    {
                        e1->rotation += TWOPI;
                    }
                }
            }
        }
        
        if(e1->mass <= 0.0f)
        {
            resetForegroundElement(e1);
        }
    }
}

void World::render(void)
{
    glClear(GL_COLOR_BUFFER_BIT);

    renderBackground();
    renderForeground();
    renderLabel();
    
    glFlush();
}

void World::handleKeypress(SDL_KeyboardEvent *event)
{
    BOOL down = (event->type == SDL_KEYDOWN);

    switch(event->keysym.sym)
    {
        case SDLK_ESCAPE:
            if(down)
            {
                m_timeToQuit = true;
            }
            break;
        
        case SDLK_TAB:
            if(down)
            {
                toggleThreadsSerialized();
            }
            break;
            
        case SDLK_f:
            if(down)
            {
                toggleFullscreen();
            }
            break;
            
        case SDLK_z:
            if(down)
            {
                adjustBackgroundCount(-BACKGROUND_STEP);
            }
            break;
            
        case SDLK_x:
            if(down)
            {
                adjustBackgroundCount(BACKGROUND_STEP);
            }
            break;
            
        case SDLK_PERIOD:
            if(down)
            {
                adjustForegroundCount(-FOREGROUND_STEP);
            }
            break;
            
        case SDLK_SLASH:
            if(down)
            {
                adjustForegroundCount(FOREGROUND_STEP);
            }
            break;
            
        default:
            break;
    }
}

BOOL World::isTimeToQuit(void)
{
    return m_timeToQuit;
}

BOOL World::areThreadsSerialized()
{
    BOOL output = true;
    
    #ifdef USE_THREADS
        // just check the update thread
        ThreadManagerSerial &tm = ThreadManagerSerial::getThreadManagerSerial();
        output = tm.isThreadSerialized(m_threadUpdateID);
    #endif
    
    return output;
}

void World::toggleThreadsSerialized()
{
    #ifdef USE_THREADS
        BOOL isSerial = areThreadsSerialized();
        
        // change all threads to opposite status
        ThreadManagerSerial &tm = ThreadManagerSerial::getThreadManagerSerial();
        tm.setThreadSerialized(m_threadUpdateID, !isSerial);
        tm.setThreadSerialized(m_threadRenderID, !isSerial);
    #endif
}

void World::toggleFullscreen()
{
    BOOL isFullscreen = m_window->isFullscreen();
    m_window->setFullscreen(!isFullscreen);
    
    setPaused(true);
}

void World::adjustBackgroundCount(int delta)
{
    // scale the background
    m_backgroundCount += delta;
    if(m_backgroundCount > BACKGROUND_MAX)
    {
        m_backgroundCount = BACKGROUND_MAX;
    }
    else if(m_backgroundCount < BACKGROUND_STEP)
    {
        m_backgroundCount = 0;
    }
    
    generateBackground();
}

void World::generateBackground(void)
{
    float extentX = m_extentXFactor * BACKGROUND_OFFSET_Z;
    float extentY = m_extentYFactor * BACKGROUND_OFFSET_Z;
    float radius = extentX / 20.0f;
    float randomX = 0.0f;
    float randomY = 0.0f;
    float randomZ = 0.0f;
    float randomAngle = 0.0f;
    float angleInc = (0.67f * PI);
    
    BackgroundElement *e = NULL;

    while(m_backgroundCount > m_backgroundMax)
    {
        e = &m_background[m_backgroundMax];
        
        // random darkish color
        e->color[0] = randomFloat(0.1f, 0.3f);
        e->color[1] = randomFloat(0.1f, 0.3f);
        e->color[2] = randomFloat(0.1f, 0.3f);
        
        // random location
        randomX = randomFloat(-(extentX * 0.5f), (extentX * 0.5f));
        randomY = randomFloat(-(extentY * 0.5f), (extentY * 0.5f));
        randomZ = -BACKGROUND_OFFSET_Z;
        
        // random orientation
        randomAngle = randomFloat(0.0f, 0.67f * PI);
        
        e->vertex[0][0] = randomX + (float)cos(randomAngle) * radius;
        e->vertex[0][1] = randomY + (float)sin(randomAngle) * radius;
        e->vertex[0][2] = randomZ;
        e->vertex[1][0] = randomX + (float)cos(randomAngle + angleInc) * radius;
        e->vertex[1][1] = randomY + (float)sin(randomAngle + angleInc) * radius;
        e->vertex[1][2] = randomZ;
        e->vertex[2][0] = randomX + (float)cos(randomAngle - angleInc) * radius;
        e->vertex[2][1] = randomY + (float)sin(randomAngle - angleInc) * radius;
        e->vertex[2][2] = randomZ;
        
        ++m_backgroundMax;
    }
}

void World::renderBackground(void)
{
    glBegin(GL_TRIANGLES);

    BackgroundElement *e = NULL;

    for(int i = 0; i < m_backgroundCount; ++i)
    {
        e = &m_background[i];
        glColor3fv(e->color);
        glVertex3fv(e->vertex[0]);
        glVertex3fv(e->vertex[1]);
        glVertex3fv(e->vertex[2]);
    }
    
    glEnd();
}

void World::adjustForegroundCount(int delta)
{
    // scale the foreground
    m_foregroundCount += delta;
    if(m_foregroundCount > FOREGROUND_MAX)
    {
        m_foregroundCount = FOREGROUND_MAX;
    }
    else if(m_foregroundCount < FOREGROUND_STEP)
    {
        m_foregroundCount = 0;
    }
    
    generateForeground();
    
    // regenerate elements if we reduce and increase again
    m_foregroundMax = m_foregroundCount;
}

void World::generateForeground(void)
{
    const float universeRadius = m_maxRadius * FOREGROUND_OFFSET_Z;
    float randomX = 0.0f;
    float randomY = 0.0f;
    float randomZ = 0.0f;
    
    ForegroundElement *e = NULL;

    while(m_foregroundCount > m_foregroundMax)
    {
        e = &m_foreground[m_foregroundMax];
        
        // random lightish color
        e->color[0] = randomFloat(0.6f, 1.0f);
        e->color[1] = randomFloat(0.6f, 1.0f);
        e->color[2] = randomFloat(0.6f, 1.0f);
        
        // random Z location
        e->location[2] = -FOREGROUND_OFFSET_Z;
        
        resetForegroundElement(e);
        
        ++m_foregroundMax;
    }
}

void World::renderForeground(void)
{
    ForegroundElement *e = NULL;

    // render space dust first
    const float spaceDustRadiusCutoff = m_foregroundRadius * 0.004f;
    const float rotationRadiusCutoff = spaceDustRadiusCutoff * 5.0f;
    glBegin(GL_POINTS);
    
    for(int i = 0; i < m_foregroundCount; ++i)
    {
        e = &m_foreground[i];
        if(e->radius < spaceDustRadiusCutoff)
        {
            glColor3fv(e->color);
            glVertex3fv(e->location);
        }
    }
    
    glEnd();
    
    // now render larger bodies
    for(int i = 0; i < m_foregroundCount; ++i)
    {
        e = &m_foreground[i];
        if(e->radius >= spaceDustRadiusCutoff)
        {
            glColor3fv(e->color);
            glPushMatrix();
            glTranslatef(e->location[0], e->location[1], e->location[2]);
            glScalef(e->radius, e->radius, e->radius);
            
            if(e->radius >= rotationRadiusCutoff)
            {
                glRotatef(((e->rotation * 180.0f) / PI), 0.0f, 0.0f, 1.0f);
            }
            
            glCallList(m_foregroundElementCallList);
            glPopMatrix();
        }
    }
}

void World::resetForegroundElement(ForegroundElement *e)
{
    float angle = randomFloat(-PI, PI);
    float distance = 0.0f;
    
    // change behavior based upon blackHoleState
    if(m_blackHoleState == BLACK_HOLE_EXPLODE)
    {
        // launch something large out of the black hole
        e->mass = randomFloat(75.0f, 150.0f);
        e->radius = radiusForMass(e->mass);
        
        float speedFactor = ((randomFloat(0.02f, 0.04f) * m_blackHoleMass) / e->mass);
        float offset = e->radius * 2.0f;
        float x = cos(angle);
        float y = sin(angle);
        
        e->location[0] = x * (m_blackHoleRadius + offset);
        e->location[1] = y * (m_blackHoleRadius + offset);
        
        e->velocity[0] = x * m_foregroundRadius * speedFactor;
        e->velocity[1] = y * m_foregroundRadius * speedFactor;
        
        m_blackHoleMass -= e->mass;
        if(m_blackHoleMass < 0.0f)
        {
            m_blackHoleMass = 0.0f;
        }
    }
    else
    {
        // space dust shows up at the galactic rim 
        distance = randomFloat(m_foregroundRadius * 0.8f, m_foregroundRadius);
        
        e->location[0] = cos(angle) * distance;
        e->location[1] = sin(angle) * distance;

        e->velocity[0] = 0.0f;
        e->velocity[1] = 0.0f;
        e->mass = 1.0f;
        e->radius = m_foregroundElementRadius;
    }
    
    e->rotation = 0.0f;
    e->rotationVelocity = randomFloat(PI * 0.1f, PI);
    if(randomFloat(-0.5f, 0.5f) > 0.0f)
    {
        e->rotationVelocity *= -1.0f;
    }
}

float World::radiusForMass(float mass)
{
    return (pow(mass, 0.3333f) * 0.004f);
}

void World::swapForegroundElement(ForegroundElement *e1, ForegroundElement *e2)
{
    ForegroundElement eSwap;
    
    size_t size = sizeof(ForegroundElement);
    
    memcpy(&eSwap, e1, size);
    memcpy(e1, e2, size);
    memcpy(e2, &eSwap, size);
}

int World::calculateTimeDeltaMilliseconds()
{
    int output = 0;
    
    m_timeUpdateCS.enter();
    
    #ifdef WIN32
        static ULARGE_INTEGER lastTime;
        
        FILETIME ft;
        ULARGE_INTEGER currentTime;
        GetSystemTimeAsFileTime(&ft);
        currentTime.HighPart = ft.dwHighDateTime;
        currentTime.LowPart = ft.dwLowDateTime;
        
        output = ((int)(currentTime.QuadPart - lastTime.QuadPart) / 10000);
        lastTime = currentTime;
    #elif LINUX
        static timeval lastTime;
        
        timeval currentTime;
        gettimeofday(&currentTime, NULL);
        
        int seconds = (currentTime.tv_sec - lastTime.tv_sec);
        int microseconds = (currentTime.tv_usec - lastTime.tv_usec);
        
        output = ((seconds * 1000) + (microseconds / 1000));
        lastTime = currentTime;
    #endif
    
    m_timeUpdateCS.leave();
    
    return output;
}

void World::updateCallsPerSecond(int milliseconds)
{
    ThreadManager &tm = ThreadManager::getThreadManager();
    
    int eventCalls = 0;
    int updateCalls = 0;
    int renderCalls = 0;
    
    eventCalls = tm.getAndResetThreadCalls(ThreadManager::INVALID_THREAD_ID);
    
    #ifdef USE_THREADS
        updateCalls = tm.getAndResetThreadCalls(m_threadUpdateID);
        renderCalls = tm.getAndResetThreadCalls(m_threadRenderID);
    #else
        updateCalls = eventCalls;
        renderCalls = eventCalls;
    #endif
    
    m_callsPerSecondCS.enter();
    m_eventCPS = ((eventCalls * 1000) / milliseconds);
    m_updateCPS = ((updateCalls * 1000) / milliseconds);
    m_renderCPS = ((renderCalls * 1000) / milliseconds);
    m_callsPerSecondCS.leave();
}

void World::renderLabel(void)
{
    BOOL isSerial = areThreadsSerialized();
    
    m_callsPerSecondCS.enter();
    SPRINTF(m_cpsBuffer, " %s -- event %d CPS -- render(%5d) %3d CPS -- update(%3d) %3d CPS ",
        (isSerial ? "serial" : "parallel"),
        m_eventCPS,
        m_backgroundCount,
        m_renderCPS,
        m_foregroundCount,
        m_updateCPS
    );
    m_callsPerSecondCS.leave();

    float scaleFactor = (LABEL_OFFSET_Z * 0.5f);
    glColor3f(1.0f, 1.0f, 1.0f);
    
    // render instructions
    glPushMatrix();
    
    glTranslatef(0.0f, ((m_extentYFactor * LABEL_OFFSET_Z) * 0.44f), -LABEL_OFFSET_Z);
    glScalef(scaleFactor, scaleFactor, 0.0f);
    
    glfStartBitmapDrawing();
	glfDrawBString(m_instructionBuffer);
	glfStopBitmapDrawing();
	
	glPopMatrix();
	
	glPushMatrix();
    
    glTranslatef(0.0f, ((m_extentYFactor * LABEL_OFFSET_Z) * 0.40f), -LABEL_OFFSET_Z);
    glScalef(scaleFactor, scaleFactor, 0.0f);
    
    glfStartBitmapDrawing();
	glfDrawBString(m_instructionBuffer2);
	glfStopBitmapDrawing();
	
	glPopMatrix();
	
	// render timing information
    glPushMatrix();
    
    glTranslatef(0.0f, -((m_extentYFactor * LABEL_OFFSET_Z) * 0.45f), -LABEL_OFFSET_Z);
    glScalef(scaleFactor, scaleFactor, 0.0f);
    
    glfStartBitmapDrawing();
	glfDrawBString(m_cpsBuffer);
	glfStopBitmapDrawing();
	
	glPopMatrix();
}
